import os
import os.path
from glob import iglob
from typing import List
from typing import Optional


def _file_generator(files_argument: List[str], is_glob: bool, filename_globs: List[str]):
    if is_glob:
        for fglob in files_argument:
            yield from iglob(fglob, recursive=True)
    else:
        for f in files_argument:
            if os.path.isdir(f):
                for g in filename_globs:
                    yield from iglob(os.path.join(f, "**", g), recursive=True)
            else:
                yield f


def file_generator(
    files_argument: List[str],
    is_glob: bool,
    filename_globs: List[str],
    ignore: Optional[List[str]] = None,
):
    gen = _file_generator(files_argument, is_glob, filename_globs)
    if ignore:
        for f in gen:
            if any(i in f for i in ignore):
                continue
            yield f
    else:
        yield from gen



import logging
from dataclasses import dataclass
from typing import Iterator
from typing import List
from typing import Optional

from ._at import EndOfString
from ._branch import Branch
from ._char import Character  # noqa: TC002
from ._repeat import InfiniteRepeat
from ._repeat import Repeat
from ._sequence import Sequence


@dataclass(frozen=True)
class Redos:
    starriness: int
    prefix_sequence: Sequence
    redos_sequence: Sequence
    repeated_character: Character
    killer: Optional[Character]

    @property
    def example_prefix(self) -> str:
        return self.prefix_sequence.example()

    def example(self, js_flavour: bool = False) -> str:
        repeated_char = self.repeated_character
        killer = self.killer
        # Try to find a repeating character which is also a killer
        if killer and (killing_repeat := repeated_char & killer):
            repeated_char = killing_repeat
            killer = None

        prefix = self.example_prefix.encode("unicode_escape").decode().replace("'", "\\'")
        repeated_char_s = (
            repeated_char.example().encode("unicode_escape").decode().replace("'", "\\'")
        )
        e = f"'{prefix}' + " if prefix else ""
        if js_flavour:
            e += f"'{repeated_char_s}'.repeat(3456)"
        else:
            e += f"'{repeated_char_s}' * 3456"

        if killer:
            killer_s = killer.example().encode("unicode_escape").decode().replace("'", "\\'")
            return e + f" + '{killer_s}'"
        return e


def find(sequence, flags: int = 0) -> List[Redos]:
    """
    Returns Redos objects sorted by severity (most starry first), then sorted by example_prefix (shortest first).
    """
    redos = []
    for r in find_redos(sequence):
        if r not in redos:
            redos.append(r)
    return sorted(redos, key=lambda r: -r.starriness * 1000 + len(r.example_prefix))


def expand_branches(seq: Sequence) -> Iterator[Sequence]:  # noqa: C901
    """
    This could blow up exponentially, but it's nicer for now to expand branches.
    """
    head = []
    for i, elem in enumerate(seq.elements):
        if isinstance(elem, Branch):
            for b in elem.get_branches():
                head_plus_branch = head + (
                    [] if not b else [b] if not isinstance(b, Sequence) else b.elements
                )
                for tail in expand_branches(Sequence(seq.elements[i + 1 :])):
                    yield Sequence(head_plus_branch + tail.elements)
            return  # All processing in yields
        elif isinstance(elem, Repeat) and elem.starriness > 10:
            logging.debug("Exponential: %s", elem)
            if isinstance(elem.repeat, (Sequence, Branch)):
                for tail in expand_branches(Sequence(seq.elements[i + 1 :])):
                    yield Sequence(head + [elem] + tail.elements)
                    for pseudo_repeat in elem.repeat.matching_repeats():
                        logging.debug("Pseudo repeat %s", pseudo_repeat)
                        yield Sequence(
                            head + [elem.alter_repeat(pseudo_repeat)] + tail.elements
                        )
            else:
                head.append(elem)
        else:
            head.append(elem)
    yield Sequence(head)


def find_redos(sequence_with_branches) -> Iterator[Redos]:
    logging.debug(sequence_with_branches)
    if not isinstance(sequence_with_branches, Sequence):  # singleton like Branch (ab|cd)
        sequence_with_branches = Sequence([sequence_with_branches])
    for seq in expand_branches(sequence_with_branches):
        yield from find_redos_in_branchless_sequence(seq)


def find_redos_in_branchless_sequence(seq: Sequence) -> Iterator[Redos]:
    logging.debug(seq)
    for i, elem in enumerate(seq.elements):
        # TODO branches  # noqa: T101
        if isinstance(elem, InfiniteRepeat) and (c := elem.overall_character_class()):
            yield from make_redos(seq, i, i + 1, c, elem.starriness)


def make_redos(  # noqa: C901
    seq: Sequence,
    sequence_start: int,
    continue_from: int,
    repeated_character: Character,
    starriness: int,
) -> Iterator[Redos]:
    # TODO branches  # noqa: T101
    character_history = [repeated_character]
    logging.debug(
        "Make ReDoS %d %d %s %d",
        sequence_start,
        continue_from,
        repeated_character,
        starriness,
    )
    for current_index in range(continue_from, len(seq)):
        elem = seq.elements[current_index]

        if isinstance(elem, EndOfString):
            # May need to go back before the matching sequence to calculate $
            elem.set_character(seq.elements[:current_index])

        eoc = elem.overall_character_class()
        new_c = repeated_character & eoc
        logging.debug("%s & %s = %s (for %s)", repeated_character, eoc, new_c, elem)

        # Handle optional elements
        if elem.minimum_length == 0:
            if elem.starriness:
                # If we have a*, we branch and try with and without it
                if new_c != repeated_character:
                    # Only branch if we have [ab]a* : if we have aa* or a[ab]* then the character class doesn't change
                    # Try without this element
                    yield from make_redos(
                        seq,
                        sequence_start,
                        current_index + 1,
                        repeated_character,
                        starriness,
                    )
            else:
                continue  # Don't care about finite repeats (abc)? or a{,4}

        # print(repeated_character, "+", elem.overall_character_class(), "->", new_c)  # noqa: E800
        if new_c is None:
            # This element will force backtracking as it's incompatible with `repeated_character`
            if elem.minimum_length and starriness > 2:
                yield redos_found(
                    seq,
                    sequence_start,
                    current_index,
                    repeated_character,
                    starriness,
                    None,
                )
            return

        starriness += elem.starriness
        repeated_character = new_c
        character_history.append(new_c)

    # Everything matched! We need to work backwards and find a 'killer' to cause backtracking if we want ReDoS
    logging.debug("Backtracking: %s", character_history)
    for current_index in reversed(range(continue_from, len(seq))):
        elem = seq.elements[current_index]
        character_history.pop()
        starriness -= elem.starriness
        if starriness <= 2:
            return
        # Can't get backtracking by not matching optional groups
        if elem.minimum_length > 0:
            # Find a character which matches the sequence and then fails on the killer
            if (match := elem.overall_character_class()) and (killer := match.negate()):
                old_repeat = character_history.pop()
                logging.debug(
                    "%s (for %s): killer=%s, repeat=%s",
                    match,
                    elem,
                    killer,
                    old_repeat,
                )
                yield redos_found(
                    seq,
                    sequence_start,
                    current_index,
                    old_repeat,
                    starriness,
                    killer,
                )
                return
    logging.debug("Backtracking: FAIL")


def redos_found(
    seq: Sequence,
    start: int,
    backtrack_at: int,
    repeated_character: Character,
    starriness: int,
    killer: Optional[Character],
) -> Redos:
    # TODO: Try to include some skipped optional parts (like `?`) just to make it nicer  # noqa: T101
    logging.debug("ReDoS found")
    return Redos(
        starriness,
        Sequence(seq.elements[:start]),
        Sequence(seq.elements[start : backtrack_at + 1]),
        repeated_character,
        killer,
    )



import logging
import string
import subprocess
import tempfile
from pathlib import Path

from packaging.version import Version

from ._util import GET_PYTHON_VERSION
from ._util import PKG_INFO
from ._util import PKG_INFO_CONFIG_REQUIRES_PYTHON
from ._util import PKG_INFO_NO_REQUIRES_PYTHON
from ._util import meson
from ._util import meson_configure
from ._util import readme_ext_to_content_type

log = logging.getLogger(__name__)


def check_pkg_info_file(config, meta):
    if 'pkg-info-file' in config:
        if not Path(config['pkg-info-file']).exists():
            builddir = tempfile.TemporaryDirectory().name
            meson_configure(builddir)
            meson('compile', '-C', builddir)
            pkg_info_file = Path(builddir) / 'PKG-INFO'
        else:
            pkg_info_file = config['pkg-info-file']
        res = '\n'.join(PKG_INFO_NO_REQUIRES_PYTHON.split('\n')[:3]).format(**meta) + '\n'
        with open(pkg_info_file, 'r') as f:
            orig_lines = f.readlines()
            for line in orig_lines:
                if (
                    line.startswith('Metadata-Version:')
                    or line.startswith('Version:')
                    or line.startswith('Name:')
                ):
                    res += config._parse_project()
                    continue
                res += line
        return res


def auto_python_version(config, python_bin: str, meta):
    python_version = Version(
        subprocess.check_output([python_bin, '-c', GET_PYTHON_VERSION])
        .decode('utf-8')
        .strip('\n')
    )
    if python_version < Version(config.min_python):
        meta.update(
            {
                'min_python': str(python_version),
                'max_python': config.max_python,
            }
        )
    elif python_version >= Version(config.max_python):
        meta.update(
            {
                'min_python': config.min_python,
                'max_python': '{}.{}'.format(
                    python_version.major, str(python_version.minor + 1)
                ),
            }
        )
    else:
        meta.update(
            {
                'min_python': config.min_python,
                'max_python': config.max_python,
            }
        )
    return meta


def check_requires_python(config, meta):
    if config['module'] == 'OZI.build':
        meta.pop('min_python')
        meta.pop('max_python')
        res = PKG_INFO_NO_REQUIRES_PYTHON.format(**meta)
    elif config.get('requires-python'):
        meta.pop('min_python')
        meta.pop('max_python')
        meta.update({'requires_python': config.get('requires-python')})
        res = PKG_INFO_CONFIG_REQUIRES_PYTHON.format(**meta)
    else:
        res = PKG_INFO.format(**meta)
    return res


def get_python_bin(config):
    option_build = config.get('meson-python-option-name')
    python = 'python3'
    if not option_build:
        log.warning(
            "meson-python-option-name not specified in the "
            + "[tool.ozi-build.metadata] section, assuming `python3`"
        )
    else:
        for opt in config.options:
            if opt['name'] == option_build:
                python = opt['value']
                break
    return python


def _parse_project_optional_dependencies(config, k: str, v: str):
    metadata = ''
    if any(
        i not in string.ascii_uppercase + string.ascii_lowercase + '-[],0123456789'
        for i in v
    ):
        raise ValueError(
            'pyproject.toml:project.optional-dependencies has invalid character in nested key "{}"'.format(
                k
            )
        )
    for j in (name for name in v.strip('[]').rstrip(',').split(',')):
        if len(j) > 0 and j[0] in string.ascii_uppercase + string.ascii_lowercase:
            for package in config.extras.get(j, []):
                metadata += 'Requires-Dist: {}; extra=="{}"\n'.format(package, k)
        else:
            raise ValueError(
                'pyproject.toml:project.optional-dependencies nested key target value "{}" invalid'.format(
                    j
                )
            )
    return metadata


def get_optional_dependencies(config):
    res = ''
    for k, v in config.extras.items():
        res += "Provides-Extra: {}\n".format(k)
        if isinstance(v, list):
            for i in v:
                if i.startswith('['):
                    res += _parse_project_optional_dependencies(config, k, i)
                else:
                    res += 'Requires-Dist: {}; extra=="{}"\n'.format(i, k)
        elif isinstance(v, str):
            res += config._parse_project_optional_dependencies(config, k, v)
            log.warning(
                'pyproject.toml:project.optional-dependencies nested key type should be a toml array, like a=["[b,c]", "[d,e]", "foo"], parsed string "{}"'.format(
                    v
                )
            )
    return res


def get_simple_headers(config):
    res = ''
    for key in [
        'summary',
        'home-page',
        'author',
        'author-email',
        'maintainer',
        'maintainer-email',
        'license',
    ]:
        if key in config:
            res += '{}: {}\n'.format(key.capitalize(), config[key])
    for key, mdata_key in [
        ('provides', 'Provides-Dist'),
        ('obsoletes', 'Obsoletes-Dist'),
        ('classifiers', 'Classifier'),
        ('project-urls', 'Project-URL'),
        ('requires-external', 'Requires-External'),
        ('dynamic', 'Dynamic'),
        ('license-file', 'License-File'),
    ]:
        vals = config.get(key, [])
        if key == 'dynamic':
            for i in vals:
                if i in {'Name', 'Version', 'Metadata-Version'}:
                    raise ValueError('{} is not a valid value for dynamic'.format(key))
        for val in vals:
            res += '{}: {}\n'.format(mdata_key, val)
    return res


def get_license_headers(config):
    res = ''
    key = 'license-expression'
    if key in config:
        if key == 'license-expression' and 'license' in config:
            raise ValueError('license and license-expression are mutually exclusive')
        header = '-'.join(map(str.capitalize, key.split('-')))
        res += '{}: {}\n'.format(header, config[key])
    return res


def get_download_url_headers(config):
    res = ''
    if 'download-url' in config:
        if '{version}' in config['download-url']:
            res += f'Download-URL: {config["download-url"].replace("{version}", config["version"])}\n'
        else:
            log.warning(
                'pyproject.toml:tools.ozi-build.metadata.download-url missing {version} replace pattern'
            )
            res += f'Download-URL: {config["download-url"]}\n'
    return res


def get_requirements_headers(config):
    res = ''
    if config.requirements:
        for package in config.requirements:
            res += 'Requires-Dist: {}\n'.format(package)
    if config.get('requires', None):
        raise ValueError(
            'pyproject.toml:tools.ozi-build.metadata.requires is deprecated as of OZI.build 1.3'
        )
    return res


def get_description_headers(config):
    res = ''
    description = ''
    description_content_type = 'text/plain'
    if 'description-file' in config:
        description_file = Path(config['description-file'])
        with open(description_file, 'r') as f:
            description = f.read()

        description_content_type = readme_ext_to_content_type.get(
            description_file.suffix.lower(), description_content_type
        )
    elif 'description' in config:
        description = config['description']

    if description:
        res += 'Description-Content-Type: {}\n'.format(description_content_type)
        res += 'Description:\n\n' + description
    return res



POLYNOMIAL_DEGREES = [
    "linear",
    "quadratic",
    "cubic",
    "quartic",
    "quintic",
    "sextic",
    "septic",
    "octic",
    "nonic",
    "decic",
]


class TextOutput:
    def __init__(self, js_flavour: bool = False):
        self.first_for_regex = True
        self.regexes = 0
        self.js_flavour = js_flavour

    def next(self):
        """Next regex being processed."""
        self.first_for_regex = True
        self.regexes += 1

    def record(self, redos, pattern, *, filename=None, lineno=None, context=None):
        if self.first_for_regex:
            if filename:
                if lineno is not None:
                    print(f"Vulnerable regex in {filename} #{lineno}")
                else:
                    print(f"Vulnerable regex in {filename}")
            print(f"Pattern: {pattern}")
            if context:
                print(f"Context: {context}")
            print("---")
            self.first_for_regex = False
        print(redos)
        stars = "\u2b50" * min(10, redos.starriness)
        degree = (
            "exponential"
            if redos.starriness > 10
            else POLYNOMIAL_DEGREES[redos.starriness - 1] if redos.starriness > 0 else "?"
        )
        print(f"Worst-case complexity: {redos.starriness} {stars} ({degree})")
        print(f"Repeated character: {redos.repeated_character}")
        if redos.killer:
            print(f"Final character to cause backtracking: {redos.killer}")
        print(f"Example: {redos.example(self.js_flavour)}\n")
